from __future__ import division

import numpy as np
import logging
import matplotlib.pyplot as plt

from sklearn.base import BaseEstimator
from sklearn.utils import check_array
from sklearn.utils import check_random_state

from dalila.utils import non_negative_projection, _check_non_negativity,\
                         _compute_clusters_and_silhouettes
from dalila.penalty import Penalty


###############################################################################
class DictionaryLearning(BaseEstimator):
    """ An estimator for dictionary learning based on alternating prox methods.

        This estimator optimises a functional of the following form:

        (1/2)||X - CD||_F^2 + phi(D) + psi(C)

        where the norm of the reconstruction error is a Frobenious matrix norm
        and the functions phi and psi are a combination of penalties that act
        row-wise on both the matrices.
        It is possible to specify which kind of penalties you want to use on
        the matrices, if there must be a non-negativity constraint and if the
        atoms in the dictionary must have norm equal to one.

        Parameters
        ----------

        k: int
            Number of atoms in which decompose the input matrix.

        dict_penalty : a sub-class of Penalty in penalty.py file, optional
            It is applied on the dictionary and it can be L0Penalty,
            L1Penalty, L2Penalty, ElasticNetPenalty

        coeff_penalty: a sub-class of Penalty class in penalty.py file,
            optional
            It is applied on the coefficients and it can be L0Penalty,
            L1Penalty, L2Penalty, ElasticNetPenalty

        dict_normalization: int, optional
            If different than zero the atoms are normalized to have l2-norm
            equal to 1.

        non_negativity: string, optional
            If 'none' (default) the atoms and the coefficients can be both
            non negative.
            If 'both' non-negativity is applied on both matrices.
            if 'coeff' non-negativity is applied only on the matrix of the
            coefficients.

        random_state: RandomState or int
            Seed to be use to initialise np.random.RandomState. If None each
            time RandomState is randomly initialised.

       """

    def __init__(self, k, dict_penalty=None, coeff_penalty=None,
                 dict_normalization=0, non_negativity='none',
                 random_state=None):

        self.k = k
        self.dict_penalty = dict_penalty
        self.coeff_penalty = coeff_penalty
        self.non_negativity = non_negativity
        self.dict_normalization = dict_normalization
        self.random_state = random_state
        self.X = None
        self.D = None
        self.C = None

    def fit(self, x, y=None, n_iter=20000, backtracking=0):

        """Function that fits the estimator on the matrix X.

        This function finds the decomposition in dictionary and coefficients
        with an alternating proximal gradient algorithm. It iteratively
        computes the gradient on one of the two matrices by keeping the other
        fixed and then apply the prox operator of the specified penalty.
        If the flat non-negative is set it project the rows of the matrices to
        the positive space, while if the dict_normalization flag is set the
        rows of the dictionaries are normalised to have l2-norm equal to 1.

        Parameters
        ----------

        x : array-like or sparse matrix shape =  (n_samples, n_features)
            The matrix to decompose.

        y:
            Inserted for compatibility with sklearn library.

        n_iter: int, optional
            Maximum number of iterations the algorithm does before stopping.

        backtracking: bool, optional
            If True a procedure of backtracking is done on the step in order
            to avoid an increasing in the objective function.

        Returns
        -------
        self : object

        """

        # ______________ parameters control________________________________#
        x = check_array(x)
        self.X = x
        n, p = x.shape
        random_state = check_random_state(self.random_state)
        self.dict_penalty = _check_penalty(self.dict_penalty)
        self.coeff_penalty = _check_penalty(self.coeff_penalty)
        _check_number_of_atoms(self.k, p, n)
        _check_non_negativity(self.non_negativity, x)

        logging.debug("Starting procedure with " +
                      str(self.k)+" number of atoms")
        logging.debug("Initializing dictionary and coefficients..")

        # ________________optimization procedure____________________________#
        self.D, self.C = \
            self._alternating_proximal_gradient_minimization(random_state,
                n_iter=n_iter, backtracking=backtracking)

        if self.dict_normalization:
            for k in range(self.k):
                normalization_factor = np.linalg.norm(self.D[k, :])
                self.D[k, :] /= normalization_factor
                self.C[:, k] *= normalization_factor

        # __________________________final controls__________________________#
        logging.debug("Finished optimization")
        if self.non_negativity == 'both':
            assert (np.min(self.D) >= 0 and np.min(self.C) >= 0)
        if self.non_negativity == 'coeff':
            assert (np.min(self.C) >= 0)

        return self

    def reconstruction_error(self):
        """
        Returns
        -------
        float:
            The reconstruction error for the current decomposition of the
            matrix. If no decomposition was run infinity is returned.
            A lower reconstruction error corresponds to a better approximation
            of the input data.
        """

        if self.X is None:
            return float("inf")
        return (np.linalg.norm(self.X - self.C.dot(self.D)) /
                np.linalg.norm(self.X))

    def objective_function_value(self, x=None, d=None, c=None):
        """

        Parameters
        ----------
        x : array-like, shape=(n_samples, n_features)
            The matrix to be decomposed.

        d: array_like, shape=(n_atoms, n_features)
            The dictionary.

        c: array-like, shape=(n_samples, n_atoms)
            The matrix of coefficients

        If one of the three is None the internal decomposition is taken, if no
        decomposition is available NaN is returned.

        Returns
        -------
        float:
            The value of the objective function.
        """
        if x is None:
            x = self.X
            if x is None:
                logging.warning('Called objective function value with no'
                                'matrices before calling fit. \n'
                                'Impossible to return an objective function '
                                'value')
                return float('NaN')

        if d is None:
            d = self.D
        if c is None:
            c = self.C

        reconstruction = np.linalg.norm(x - c.dot(d))**2
        penalty_dictionary = self.dict_penalty.value(d)
        penalty_coefficient = self.coeff_penalty.value(c)
        return reconstruction + penalty_coefficient + penalty_dictionary

    def decomposition(self):
        """
        Returns
        -------

        C: array-like, shape = (n_samples, k)
            The learnt coefficients.

        D: array_like, shape = (k, n_features)
            The learnt dictionary.

        """
        return self.C, self.D

    def score(self, *args):
        """
        Parameters
        ----------

        *args: optional
            Introduced for compatibility with sklearn GridSearchCV

        Returns
        -------
        float
            The score of the current decomposition. BIC value computed as
            -log(k)*log(n_samples) - 2.3*(self.objective_function_value())
            the highest is the score and better the decomposition is.
        """

        if self.X is None:
            return float("-inf")
        return - (np.log(self.X.shape[0])*self.k \
                    + 2 * np.linalg.norm(self.X - self.C.dot(self.D))**2)

    def _alternating_proximal_gradient_minimization(self, random_state,
                                                    n_iter=20000,
                                                    backtracking=0):
        x = self.X
        n, p = x.shape
        d = non_negative_projection(random_state.rand(self.k, p)*10-5,
                                    self.non_negativity, 'dict')
        c = non_negative_projection(random_state.rand(n, self.k)*10-5,
                                    self.non_negativity, 'coeff')

        gamma_c = 1.1
        gamma_d = gamma_c
        step_d, step_c = _step_lipschitz(d, c,
                                         gamma_d=gamma_d, gamma_c=gamma_c)
        epsilon = 1e-4
        objective = self.objective_function_value(d=d, c=c)

        d_old = d
        c_old = c
        logging.debug("Starting optimization")
        for i in range(n_iter):
            gradient_d = c.T.dot(c.dot(d) - x)
            gradient_c = (c.dot(d) - x).dot(d.T)

            if backtracking:
                d, c = self._update_with_backtracking(d, c, gradient_d,
                                                      gradient_c, step_d,
                                                      step_c)
            else:
                d, c = self._simple_update(d, c, gradient_d, gradient_c,
                                           step_d, step_c)

            new_objective = self.objective_function_value(d=d, c=c)
            difference_objective = new_objective - objective
            objective = new_objective
            difference_d = np.linalg.norm(d - d_old)
            difference_c = np.linalg.norm(c - c_old)
            d_old = d
            c_old = c

            step_d, step_c = _step_lipschitz(d, c,
                                             gamma_c=gamma_c, gamma_d=gamma_d)

            logging.debug("Iteration: "+str(i))
            logging.debug("Objective function: "+str(objective))
            logging.debug("Difference between new objective and "
                          "previous one: "+str(difference_objective))
            logging.debug("Difference between previous and new dictionary: " +
                          str(difference_d))
            logging.debug("Difference between previous and new coefficients:" +
                          str(difference_c)+"`\n\n")

            assert ((not np.isnan(difference_objective)) and
                    (not np.isinf(difference_objective)) and
                    (abs(difference_objective) < 1e+20))

            if (abs(difference_objective) <= epsilon and
                    difference_d <= epsilon and
                    difference_c <= epsilon):
                break

        return d, c

    def _simple_update(self, d, c, gradient_d, gradient_c,
                       step_d, step_c):
        d = self.dict_penalty. \
            apply_prox_operator(d - step_d * gradient_d, gamma=step_d)
        d = non_negative_projection(d, self.non_negativity, 'dict')

        c = self.coeff_penalty. \
            apply_prox_operator(c - step_c * gradient_c, gamma=step_c)
        c = non_negative_projection(c, self.non_negativity, 'coeff')
        return d, c

    def _update_with_backtracking(self, d, c, gradient_d, gradient_c,
                                  step_d, step_c):
        x = self.X
        d_0 = d
        c_0 = c
        error = 0.5 * np.linalg.norm(x - c_0.dot(d_0))**2
        sigma = 0.9
        for i in range(1, 1000):
            # compute new matrices
            d_1 = self.dict_penalty. \
                apply_prox_operator(d_0 - step_d * gradient_d, gamma=step_d)
            d_1 = non_negative_projection(d_1, self.non_negativity, 'dict')
            difference_d = np.linalg.norm(d_1 - d_0)

            c_1 = self.coeff_penalty. \
                apply_prox_operator(c_0 - step_c * gradient_c, gamma=step_c)
            c_1 = non_negative_projection(c_1, self.non_negativity, 'coeff')
            difference_c = np.linalg.norm(c_1 - c_0)

            first_part = self.objective_function_value(d=d_1, c=c_1)

            second_part = (error +
                           self.dict_penalty.value(d_1) +
                           self.coeff_penalty.value(c_1) +
                           (difference_c*gradient_c).trace() +
                           (difference_d*gradient_d).trace() +
                           step_c*sigma**i * difference_c +
                           step_d*sigma**i * difference_d)

            if first_part <= second_part:
                break

        return d_1, c_1


###############################################################################
class StabilityDictionaryLearning(DictionaryLearning):
    """ Estimator for dictionary learning (DL) based on prox methods and clusters.

        This estimator optimises a functional of the following form:

        (1/2)||X - CD||_F^2 + phi(D) + psi(C)

        where the norm of the reconstruction error is a Frobenious matrix norm
        and the functions phi and psi are a combination of penalties that act
        row-wise on both the matrices.
        It is possible to specify which kind of penalties you want to use on
        the matrices, if there must be a non-negativity constraint and if the
        atoms in the dictionary must have norm equal to one.

        The algorithm performs many decomposition using DL and after each
        iteration it clusters the atoms obtained in every iteration in order to
        have final atoms that are stable w.r.t to data noise.

        For more details see Alexandrov et al, Cell 2013.

        Parameters
        ----------

        k: int
            Number of atoms in which decompose the input matrix.

        dict_penalty : a sub-class of Penalty in penalty.py file,
            optional
            It is applied on the dictionary and it can be L0Penalty,
            L1Penalty, L2Penalty, ElasticNetPenalty

        coeff_penalty: sa sub-class of Penalty class in penalty.py file,
            optional
            It is applied on the coefficients and it can be L0Penalty,
            L1Penalty, L2Penalty, ElasticNetPenalty

        dict_normalization: int, optional
            If different than zero the atoms are normalized to have l2-norm
            equal to 1.

        non_negativity: string, optional
            If 'none' (default) the atoms and the coefficients can be both
            non negative.
            If 'both' non-negativity is applied on both matrices.
            if 'coeff' non-negativity is applied only on the matrix of the
            coefficients.


        random_state: RandomState or int
            Seed to be use to initialise np.random.RandomState. If None each
            time RandomState is randomly initialised.

       """

    def __init__(self, k, dict_penalty=None, coeff_penalty=None,
                 dict_normalization=1, non_negativity='none',
                 random_state=None):
        super(StabilityDictionaryLearning, self)\
            .__init__(k, dict_penalty, coeff_penalty, 1, non_negativity,
                      random_state)
        self.meanD = None
        self.meanC = None
        self.stability = None
        self.Ds_sequence = None
        self.Cs_sequence = None

    def fit(self, x, y=None, backtracking=0, n_iter=20000, epsilon=1e-4):
        """Function that fits the estimator on the matrix X.

           This function finds the decomposition in dictionary and coefficients
           with an alternating proximal gradient algorithm. It iteratively
           computes the gradient on one of the two matrices by keeping the
           other fixed and then apply the prox operator of the specified
           penalty. If the flat non-negative is set it project the rows of the
           matrices to the positive space, while if the dict_normalization flag
           is set the rows of the dictionaries are normalised to have l2-norm
           equal to 1.

           Parameters
           ----------

           x : array-like or sparse matrix shape =  (n_samples, n_features)
               The matrix to decompose.

           y:
               Inserted for compatibility with sklearn library.

           n_iter: int, optional
               Maximum number of iterations the algorithm does before stopping.

           backtracking: bool, optional
               If True a procedure of backtracking is done on the step in order
               to avoid an increasing in the objective function.

           epsilon: float, optional
                The difference between the each iteration of DL at which the
                algorithm stops to return the solution.

           Returns
           -------
           self : object

           """
        x = check_array(x)
        self.X = x
        n, p = x.shape
        random_state = check_random_state(self.random_state)
        self.dict_penalty = _check_penalty(self.dict_penalty)
        self.coeff_penalty = _check_penalty(self.coeff_penalty)
        _check_number_of_atoms(self.k, p, n)
        _check_non_negativity(self.non_negativity, x)

        difference = 10
        n, d = x.shape

        D = random_state.rand(self.k, d)
        Ds = []
        Cs = []

        mean_D = D
        # until convergence
        while difference > epsilon:
            old_mean_D = mean_D
            # random bootstrap with replacement and fit
            samples = _sampling(x, random_state)
            super(StabilityDictionaryLearning, self)\
                .fit(samples, n_iter=n_iter, backtracking=backtracking)
            Ds.append(self.D)
            Cs.append(self.C)

            Ds, Cs, mean_D, mean_C, stability = \
                _compute_clusters_and_silhouettes(Ds, Cs)

            # normalization
            for i in range(mean_D.shape[0]):
                mean_D[i, :] = mean_D[i, :] / np.sum(mean_D[i, :])

            difference = np.sum((old_mean_D - mean_D) ** 2)
            logging.debug("Difference " + str(difference))

        self.meanD = mean_D
        self.meanC = mean_C
        self.stability = stability
        self.Ds_sequence = Ds
        self.Cs_sequence = Cs
        return self

    def decomposition(self):
        """
            Returns
            -------

            C: array-like, shape = (n_samples, k)
                The learnt mean coefficients.

            D: array_like, shape = (k, n_features)
                The learnt mean dictionary, the centroids of the clusters of
                atoms.

            """
        return self.meanC, self.meanD

    def stability(self):
        """

        Returns
        -------
        float
            The mean stability obtained with the clusterization. Where for
            stability we mean the average silhouette over all the clusters.
        """
        return self.stability


# ____________________________UTILITY FUNCTIONS_______________________________
def _step_lipschitz(d, c, gamma_d,  gamma_c):
    step_c = max(0.0001, gamma_c * np.linalg.norm(d.T.dot(d)))
    step_d = max(0.0001, gamma_d * np.linalg.norm(c.dot(c.T)))
    return (1/step_d), (1/step_c)


def _check_penalty(penalty):

    if penalty is None:
        return Penalty()

    if not isinstance(penalty, Penalty):
        logging.warning('The penalty is not a subclass of the '
                        'right type.')
        raise TypeError('The penalty is not a subclass of the '
                        'right type.')

    return penalty


def _check_number_of_atoms(k, p, n):
    if k is None:
        logging.warning('No number of atoms was given. '
                        'Impossible to do optimization.')
        raise ValueError('No number of atoms was given. '
                        'Impossible to do optimization.')
    if k <= 0 or k > min(n, p):
        logging.warning('The number of atoms must be greater than zero '
                        'and less than the min of the dimensions of X.')
        raise ValueError('The number of atoms must be greater than zero '
                        'and less than the min of the dimensions of X.')


def _sampling(X, random_state):  # sampling with replacement
    selected = random_state.randint(0, high=X.shape[0],
                                    size=(X.shape[0]))
    return X[selected, :]





from __future__ import division

import logging

from sklearn.externals.joblib import Parallel, parallel_backend, \
    register_parallel_backend
from distributed.joblib import DistributedBackend
from sklearn.model_selection import GridSearchCV, ShuffleSplit
from sklearn.utils import check_array, check_random_state
from multiprocessing import cpu_count

from dalila.dictionary_learning import DictionaryLearning
from dalila.representation_learning import RepresentationLearning
from dalila.utils import _check_non_negativity, MonteCarloBootstrap


def tune_parameters_DL(X, estimator=None, analysis=3, non_negative="none",
                       distributed=0, scheduler_host="", range_k=None,
                       dict_penalty_range=(0.0001, 1, 10),
                       coeff_penalty_range=(0.0001, 1, 10),
                       fit_params = {},
                       scoring_function=None,
                       random_state=None):
    """
    Parameters tuner.

    It tunes the parameters of a dictionary learning estimator using 3-splits
    monte carlo sampling cross validation.

    Parameters
    ----------
    X: array-like, shape=(n_samples, n_features)
        The matrix to decompose and analyse.

    estimator: DictionaryLearning class, optional
        The estimator you want to use to analyse the matrix. If None only the
        research on the best number of atoms will be done.

    analysis: int, optional
        The type of tuning you want to perform.
        - 0: tune together number of atoms and dictionary penalty and then the
             coefficients penalty
        - 1: tune only the penalties and take the number of atoms as specified
             in the estimator
        - 2: tune only the number of atoms
        - 3: tune all together, number of atoms and penalties

    non_negative: string, optional
        If "none" no negativity is imposed on the decomposition, if "coeff"
        only negativity on the coefficient is imposed. If "both" negativiy is
        on both decomposition matrices.

    distributed: int, optional
        If 0 the parameters research will be executed in parallel on the
        computer the script is launched.
        If 1 the parameters research will be executed sequentially.
        If 2 the parameters research will be distributed on multiple machines
        connected by dask. In this case also scheduler_host must be speficied.

    scheduler_host: string, optional
        If distributed=2 it is necessary to specify the scheduler of the dask
        network. The string must be "ip_address:port", for example:
        "10.251.61.226:8786"

    range_k: int or list, optional
        The maximum number of atoms to try when you search for the right k or
        the list of possible values to try.
        If None range_k will be computed as int(min(p, 0.75 * n) / 2)

    dict_penalty_range: float tuple, optional (low, high, number)
        It gives the interval in which tune the dictionary penalty and the
        number of values to try.

    coeff_penalty_range: float tuple, optional (low, high, number)
        It gives the interval in which tune the coefficient penalty and the
        number of values to try.

    fit_params: dictionary, optional
        The parameters to pass to the fitting procedure during GridSearch.

    scoring_function: callable or None, default=None
        A scorer callable object / function with signature
        scorer(estimator, X, y=None). If None, the score method of the
        estimator is used.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    GridSearchCV
    The resulting GridSearch.

    """

    # ------------------parameters control ---------------------------------- #
    X = check_array(X)
    random_state = check_random_state(random_state)
    _check_range(dict_penalty_range)
    _check_range(coeff_penalty_range)
    _check_non_negativity(non_negative, X)

    if estimator is None:
        analysis = 2
    else:
        _check_estimator(estimator)
        if estimator.non_negativity == "none":
            estimator.non_negativity = non_negative

    n, p = X.shape
    if range_k is None:
        range_k = int(min(p, 0.75 * n) / 2)  # generally the optimal
                                           # number of k is low

    if (analysis in [0, 1, 3] and
       (dict_penalty_range is None or coeff_penalty_range is None)):
        logging.ERROR("The range cannot be None")
        sys.exit(0)

    if distributed == 2:
        if scheduler_host is None:
            logging.ERROR("Distributed execution requires a scheduler "
                          "specification. Changing the type to parallel.")
            distributed = 1
        distributed = _check_scheduler(scheduler_host)

    # find first the paramaters on the dictionary and after the coefficients
    if analysis == 0:
        params = _get_params_dict(estimator,
                                  dict_penalty_range=dict_penalty_range)
        if type(range_k) is int:
            params['k'] = list(range(2, range_k))
        else:
            params['k'] = range_k

        jobs = 1 if distributed == 1 else cpu_count()
        gscv = GridSearchCV(estimator, params, cv=ss, n_jobs=jobs,
                            scoring=scoring_function,
                            iid=True, refit=True, verbose=1)
        if distributed == 2:
            register_parallel_backend('distributed', DistributedBackend)
            with parallel_backend('distributed', scheduler_host=scheduler_host):
                gscv.fit(X)
        else:
            gscv.fit(X)
        estimator = gscv.best_estimator_
        params = _get_params_coeff(estimator, coeff_penalty_range)
    # find only the penalties together
    elif analysis == 1:
        params = _get_params(estimator, dict_penalty_range,
                             coeff_penalty_range)
    # find only the number of atoms
    elif analysis == 2:
        if type(range_k) is int:
            params = {'k': list(range(2, max_k))}
        else:
            params = {'k': range_k}
    # find everything together
    elif analysis == 3:
        params = _get_params(estimator, dict_penalty_range,
                             coeff_penalty_range)

        if type(range_k) is int:
            params['k'] = list(range(2, range_k))
        else:
            params['k'] = range_k
    else:
        logging.error("Unknown type of research, please try with another "
                      "setting")
        raise ValueError("Unkown type of research, please try with another"
                         "setting")

    ss = MonteCarloBootstrap(n_splits=3, test_size=0.1,
                             random_state=random_state)
    jobs = 1 if distributed == 1 else cpu_count()
    gscv = GridSearchCV(estimator, params, cv=ss, fit_params=fit_params,
                        n_jobs=jobs, iid=True, scoring=scoring_function,
                        refit=True, verbose=1)
    if distributed == 2:
        register_parallel_backend('distributed', DistributedBackend)
        with parallel_backend('distributed',
                              scheduler_host=scheduler_host):
            gscv.fit(X)
    else:
        gscv.fit(X)
    return gscv


def tune_parameters_RL(X, estimator, non_negative=0,  distributed=0,
                       scheduler_host="", coeff_penalty_range=(0.0001, 1, 10),
                       fit_params={}, scoring_function=None,
                       random_state=None):
    """
    Parameters tuner.

    It tunes the parameters of a representations learning estimator using
    3-splits monte carlo sampling cross validation.

    Parameters
    ----------
    X: array-like, shape=(n_samples, n_features)
        The matrix to decompose and analyse.

    D: array-like, shape=(n_atoms, n_features)
        The dictionary.

    estimator: RepresentationLearning class, optional
        The estimator you want to use to analyse the matrix.

    non_negative: boolean, optional

    distributed: int, optional
        If 0 the parameters research will be executed in parallel on the
        computer the script is launched.
        If 1 the parameters research will be executed sequentially.
        If 2 the parameters research will be distributed on multiple machines
        connected by dask. In this case also scheduler_host must be speficied.

    scheduler_host: string, optional
        If distributed=2 it is necessary to specify the scheduler of the dask
        network. The string must be "ip_address:port", for example:
        "10.251.61.226:8786"

    coeff_penalty_range: float tuple, optional (low, high, number)
        It gives the interval in which tune the coefficient penalty and the
        number of values to try.

    fit_params: dictionary, optional
        The parameters to pass to the fitting procedure during GridSearch.

    scoring_function: callable or None, default=None
        A scorer callable object / function with signature
        scorer(estimator, X, y=None). If None, the score method of the
        estimator is used.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    GridSearchCV
    The resulting GridSearch.

    """

    # ------------------parameters control ---------------------------------- #
    X = check_array(X)
    random_state = check_random_state(random_state)
    _check_range(coeff_penalty_range)
    if estimator is None:
        logging.error("passed estimator was None")
        raise ValueError("passed estimator was None")
    _check_estimator(estimator)

    estimator.non_negativity = non_negative

    if distributed == 2:
        if scheduler_host is None:
            logging.ERROR("Distributed execution requires a scheduler "
                          "specification. Changing the type to parallel.")
            distributed = 1
        distributed = _check_scheduler(scheduler_host)

    ss = MonteCarloBootstrap(n_splits=3, test_size=0.1,
                                 random_state=random_state)

    params = _get_params_coeff(estimator, coeff_penalty_range,
                               representation_learning=1)

    jobs = 1 if distributed == 1 else cpu_count()
    gscv = GridSearchCV(estimator, params, cv=ss, n_jobs=(cpu_count() - 5),
                        fit_params=fit_params, iid=True, refit=True,
                        scoring=scoring_function, verbose=1)
    if distributed == 2:
        register_parallel_backend('distributed', DistributedBackend)
        with parallel_backend('distributed',
                              scheduler_host=scheduler_host):
            gscv.fit(X)
    else:
        gscv.fit(X)

    return gscv


def _get_params_dict(estimator, dict_penalty_range):
    if estimator.dict_penalty is None:
        return {}

    return {'dict_penalty': (estimator.dict_penalty.
                                 make_grid(dict_penalty_range[0],
                                           dict_penalty_range[1],
                                           dict_penalty_range[2]))}


def _get_params_coeff(estimator, coeff_penalty_range,
                      representation_learning=0):
    if representation_learning:
        if estimator.penalty is None:
            return {}
    else:
        if estimator.coeff_penalty is None:
            return {}

    if representation_learning:
        return {'penalty': (estimator.penalty.
                             make_grid(coeff_penalty_range[0],
                                       coeff_penalty_range[1],
                                       coeff_penalty_range[2]))}

    return {'coeff_penalty': (estimator.coeff_penalty.
                             make_grid(coeff_penalty_range[0],
                                       coeff_penalty_range[1],
                                       coeff_penalty_range[2]))}


def _get_params(estimator, dict_penalty_range, coeff_penalty_range):
    if estimator.dict_penalty is None and estimator.coeff_penalty is None:
        return {}

    if estimator.dict_penalty is None:
        if coeff_penalty_range is None:
            return {}
        else:
            return {'coeff_penalty': (estimator.coeff_penalty.
                                      make_grid(coeff_penalty_range[0],
                                                coeff_penalty_range[1],
                                                coeff_penalty_range[2]))}
    if estimator.coeff_penalty is None:
        if dict_penalty_range is None:
            return {}
        else:
            return {'dict_penalty': (estimator.dict_penalty.
                                     make_grid(dict_penalty_range[0],
                                               dict_penalty_range[1],
                                               dict_penalty_range[2]))}
    return {'dict_penalty': (estimator.dict_penalty.
                             make_grid(dict_penalty_range[0],
                                       dict_penalty_range[1],
                                       dict_penalty_range[2])),
            'coeff_penalty': (estimator.coeff_penalty.
                              make_grid(coeff_penalty_range[0],
                                        coeff_penalty_range[1],
                                        coeff_penalty_range[2]))}


def _check_estimator(estimator):
    if not ((isinstance(estimator, DictionaryLearning)) or
            isinstance(estimator, RepresentationLearning)):
        logging.error('Unknown estimator.')
        raise TypeError('Unknown estimator.')


def _check_range(r):
    if r is None:
        return
    if not len(r) == 3:
        logging.error('Too few elements in range specification.')
        raise ValueError('Too few elements in range specification.')

    if r[0] < 0:
        logging.error('The minimum number for the range is 0.')
        raise ValueError('The minimum number for the range is 0.')

    if r[2] > 10:
        logging.warning('Using more than ten values for each penalty may '
                        'require a long computational time.')


def _check_scheduler(s):
    a = s.split(".")
    l = a[-1].split(":")
    IP = a[0:3] + [l[0]]
    port = l[1]

    accepted = True
    if len(IP) != 4:
        accepted = False
    for x in IP:
        if not x.isdigit():
            accepted = False
        i = int(x)
        if i < 0 or i > 255:
            accepted = False
    if not accepted:
        logging.warning("The given IP does not respect the requirements."
                        "Parallel version will be executed")
        return 0
    return 2



from __future__ import print_function, division

import logging

import numpy as np
from itertools import product, combinations, chain
import bintrees


class Penalty:
    """
    Super class that represents a general empty penalty.
    """

    def apply_by_row(self, x, gamma):
        if x.ndim < 2:
            return self.prox_operator(x, gamma)
        else:
            new_x = np.zeros_like(x)
            for r in range(0, x.shape[0]):
                new_x[r, :] = self.prox_operator(x[r, :], gamma)
        return new_x

    def apply_by_col(self, x, gamma):
        if x.ndim < 2:
            return self.prox_operator(x, gamma)
        else:
            new_x = np.zeros_like(x)
            for c in range(0, x.shape[1]):
                new_x[:, c] = self.prox_operator(x[:, c], gamma)
            return new_x

    def apply_prox_operator(self, x, gamma):
        return x

    def prox_operator(self, x, gamma):
        return x

    def make_grid(self):
        return []

    def value(self, x):
        return 0


class L1Penalty(Penalty):

    """
    Class representing l1 penalty.

    The prox operator in case of 2D matrix will be applied rowwise.

    Parameters
    ----------

    _lambda: float
    The regularization value. Allawed values are higher than 0.

    """

    def __init__(self, _lambda):
       self._lambda = _lambda

    def apply_prox_operator(self, x, gamma):
        if self._lambda < 0:
            logging.error("A negative regularization parameter was used")
            raise ValueError("A negative regularization parameter was used")

        return self.apply_by_row(x, gamma)

    def prox_operator(self, x, gamma):
        sign = np.sign(x)
        np.maximum(np.abs(x) - (self._lambda*gamma), 0, out=x)
        x *= sign
        return x

    def make_grid(self, low=0.001, high=1, number=10):
        values = np.logspace(np.log10(low), np.log10(high), number)
        l = []
        for (i, v) in enumerate(values):
           l.append(L1Penalty(v))
        return l

    def __str__(self):
        return "L1Penalty(" + str(self._lambda) + ")"

    def value(self, x):
        return self._lambda * np.linalg.norm(x, 1)


class L2Penalty(Penalty):
    """
        Class representing l2 penalty.

        The prox operator in case of 2D matrix will be applied rowwise.

        Parameters
        ----------

        _lambda: float
        The regularization value. Allowed values are higher than 0.

        """

    def __init__(self, _lambda):
         self._lambda = _lambda

    def apply_prox_operator(self, x, gamma):
        if self._lambda < 0:
            logging.error("A negative regularisation parameter was used")
            raise ValueError("A negative regularization parameter was used")

        return self.apply_by_row(x, gamma)

    def prox_operator(self, x, gamma):
        norm = np.linalg.norm(x) + 1e-10  # added constant for stability
        x *= max(1 - (gamma*self._lambda) / norm, 0)
        return x

    def make_grid(self, low=0.001, high=1, number=10):
        values = np.logspace(np.log10(low), np.log10(high), number)
        l = []
        for (i, v) in enumerate(values):
            l.append(L2Penalty(v))
        return l

    def __str__(self):
        return "L2Penalty(" + str(self._lambda) + ")"

    def value(self, x):
        return self._lambda * np.sum(np.apply_along_axis(np.linalg.norm,
                                                         axis=1, arr=x))


class ElasticNetPenalty(Penalty):
    """
        Class representing elastic net penalty.

        The prox operator in case of 2D matrix will be applied rowwise.

        Parameters
        ----------

        _lambda1: float
        The regularization value for the l1 penalty.

        _lambda2: float
        The regularization value for the l2 penalty.

        alpha: float
            alpha is the percentage of l1 regularization while (1-alpha) is the
            percentage of l2 regularization.

        """

    def __init__(self, _lambda1, _lambda2, alpha):
        self._lambda1 = _lambda1
        self._lambda2 = _lambda2
        self.alpha = alpha

    def apply_prox_operator(self, x, gamma):
        if self._lambda1 < 0 or self._lambda2 < 0:
            logging.error("A negative regularisation parameter was used")
            raise ValueError("A negative regularization parameter was used")
        if self.alpha < 0 or self.alpha>1:
            logging.error("The alpha value of elastic net penalty has to be "
                          "in the interval [0,1]")
            raise ValueError("The alpha value of elastic net penalty has to be "
                          "in the interval [0,1]")

        return self.apply_by_row(x, gamma)

    def prox_operator(self, x, gamma):
        sign = np.sign(x)
        np.maximum(np.abs(x) - (self._lambda1 * gamma), 0, out=x)
        x *= sign
        x /= (1. + self.alpha * self._lambda2 * gamma)
        return x

    def make_grid(self, low=0.001, high=1, number=5):
        a = np.logspace(np.log10(low), np.log10(high), number)
        b = np.logspace(np.log10(low), np.log10(high), number)
        alpha = np.arange(0.1, 1, 0.2)
        l1 = product(a, b)
        values = product(l1, alpha)
        l = []
        for (i, v) in enumerate(values):
            l.append(ElasticNetPenalty(v[0][0], v[0][1], v[1]))
        return l

    def __str__(self):
        return "ElasticNetPenalty( " \
               + str(self._lambda1) + ", " \
               + str(self._lambda2) + ", " + str(self.alpha) + ")"

    def value(self, x):
        l1 = self._lambda1*self.alpha * np.linalg.norm(x, 1)
        l2 = self._lambda2*(1-self.alpha) + \
             np.sum(np.apply_along_axis(np.linalg.norm, axis=1, arr=x))
        return l1 + l2


class L0Penalty(Penalty):
    """
        Class representing L0 penalty.

        The prox operator in case of 2D matrix will be applied rowwise.

        Parameters
        ----------

        _s: int
        The number of non-zero elements in each row.

        """

    def __init__(self, s):
        self.s = s

    def apply_prox_operator(self, x, gamma):
        if self.s < 0:
            logging.error("A negative regularisation parameter was used")
            raise ValueError("A negative regularization parameter was used")
        if self.s > x.shape[1]:
            logging.error("The number of non-zero elements to impose with L0 "
                          "penalty cannot be higher than the number of "
                          "features")
            raise ValueError("The number of non-zero elements to impose with "
                             "L0 penalty cannot be higher than the number of "
                             "features")
        return self.apply_by_row(x, gamma)

    def prox_operator(self, x, gamma):
        indices = np.argsort(x)
        x_new = np.zeros_like(x)
        x_new[indices[-self.s:]] = x[indices[-self.s:]]
        return x_new

    def make_grid(self, low=0, high=1, number=5):
        values = np.linspace(low, high, number)
        l = [L0Penalty(v) for v in values]
        return l

    def __str__(self):
        return "L0Penalty( " \
               + str(self.s) + ")"

    def value(self, x):
        return self.s


class GroupLassoPenalty(Penalty):
    """
        Class representing non-overlapping Group Lasso penalty.

        The prox operator in case of 2D matrix will be applied rowwise.

        Parameters
        ----------

        _groups: list of int
        List of the groups in which subdivide the considered feature space.
        There cannot be overlaps.

        _lambda: float
        The regularization value. Allawed values are higher than 0.

        """

    def __init__(self, _groups, _lambda):
        self._groups = _groups
        self._lambda = _lambda

    def apply_prox_operator(self, x, gamma):
        if self._lambda < 0:
            logging.error("A negative regularisation parameter was used")
            raise ValueError("A negative regularization parameter was used")

        l = list(set().union(*self._groups))
        if not (l == list(np.arange(x.shape[1]))):
            logging.error("The groups in group lasso must cover all the "
                          "features")
            raise ValueError("The groups in group lasso must cover all the "
                          "features")

        for pair in combinations(self._groups, r=2):
            if len(set(pair[0]) & set(pair[1])) > 0:
                logging.error("There are overlapping groups")
                raise ValueError("There are overlapping groups")

        new_x = np.zeros_like(x)
        for r in range(0, x.shape[0]):
            for g in self._groups:
                new_x[r, g] = self.prox_operator(x[r, g], gamma)
        return new_x

    def prox_operator(self, x, gamma):
        if np.linalg.norm(x) < self._lambda*gamma:
            return np.zeros_like(x)
        x *= (1 - (self._lambda*gamma)/np.linalg.norm(x))
        return x

    def make_grid(self, low=0, high=1, number=5):
        values = np.linspace(low, high, number)
        l = [GroupLassoPenalty(self._groups, v) for v in values]
        return l

    def __str__(self):
        return "GroupLassoPenalty( " \
               + str(self._groups) + str(self._lambda) + ")"

    def value(self, x):
        res = 0
        for r in range(x.shape[0]):
            for g in self._groups:
                res += np.linalg.norm(x[r, g])
        return self._lambda * res


class LInfPenalty(Penalty):

    def __init__(self, _lambda):
        self._lambda = _lambda

    def apply_prox_operator(self, x, gamma):
        if self._lambda < 0:
            logging.error("A negative regularisation parameter was used")
            raise ValueError("A negative regularization parameter was used")
        return self.apply_by_col(x, gamma)

    def prox_operator(self, x, gamma):
        # norm = np.linalg.norm(x) + 1e-10  # added constant for stability
        # x *= max(1 - (gamma*self._lambda) / norm, 0)
        x -= gamma*self._lambda * \
             self._projection_L1ball(x/(gamma*self._lambda))
        return x

    def make_grid(self, low=-3, high=1, number=10):
        values = np.logspace(low, high, number)
        l = []
        for (i, v) in enumerate(values):
            l.append(LInfPenalty(v))
        return l

    def projection_L1ball(self, v):

        """Find the projection of a vector onto the L1 ball. See for reference
        'Efficient projections onto L1-ball for learning in high dimensions'
        Duchi , Schwartz, Singer. SECTION 4, Figure 2.
        Parameters
        -------------------
        v : vector to be projected
        Returns
        -------------------
        w : projection on v onto the L1 ball of radius z
        """

        vector_copy = np.abs(v)
        keys_ar = np.unique(vector_copy)
        r = map(lambda k: (np.where(k == vector_copy)[0]).sum(), keys_ar)
        r = list(r)
        # r = np.where(j == vector_copy)[0].sum()
        # print("where ", r)
        # r = map(lambda k: len(np.where(k == vector_copy)[0]), keys_ar)
        # print("where2 ", r)
        n_elements = np.cumsum(r[::-1])[::-1]  # r di cui abbiamo bisogno
        s_sum = keys_ar * r
        cum_sum = np.cumsum(s_sum[::-1])[::-1]

        dictionary = {}
        for i in range(len(keys_ar)):
            dictionary[keys_ar[i]] = [n_elements[i], cum_sum[i]]

        rb_tree = bintrees.RBTree(dictionary)
        theta = self._pivotsearch(rb_tree, rb_tree._root, 0., 0.)

        return np.clip(vector_copy - theta, a_min=0, a_max=np.inf) * np.sign(v)

    def _pivotsearch(self, rb_tree, v, rho, s, v_star=np.inf, rho_star=0.,
                     s_star=0.):
        z = 1.  # ray of the ball, 1
        rho_hat = v.value[0]
        s_hat = v.value[1]

        if s_hat < v.key * rho_hat + z:
            if v_star > v.key:
                v_star = v.key
                rho_star = rho_hat
                s_star = s_hat
            if self._is_leaf(v):
                return (s_star - z) / rho_star
            if v.left is not None:
                del rb_tree[v.key:]
                return self._pivotsearch(rb_tree, rb_tree._root, rho_hat,
                                         s_hat, v_star, rho_star,
                                         s_star)  # node v.left
            else:
                return (s_star - z) / rho_star  # "no left child"

        else:
            if self._is_leaf(v):
                return (s_star - z) / rho_star
            if v.right is not None:
                del rb_tree[:v.key]
                del rb_tree[v.key]
                return self._pivotsearch(rb_tree, rb_tree._root, rho, s,
                                         v_star, rho_star,
                                         s_star)  # node v.right
            else:
                return (s_star - z) / rho_star  # "no right child"

    def _is_leaf(self, v):
        return v.left is None and v.right is None





from __future__ import print_function, division

import numpy as np
import logging

from sklearn.base import BaseEstimator
from sklearn.utils import check_array
from sklearn.utils import check_random_state

from dalila.utils import non_negative_projection
from dalila.dictionary_learning import _check_penalty


class RepresentationLearning(BaseEstimator):
    """ An estimator for finding coefficients given the dictionary.

        This estimator, given X and D, optimises a functional of the
        following form:

        (1/2)||X - CD||_F^2 +  psi(C)

        where the norm of the reconstruction error is a Frobenious matrix norm
        and the function psi is a combination of penalties that act
        row-wise on both the matrices.
        It is possible to specify which kind of penalties you want to use on
        the matrices, if there must be a non-negativity constraint and if the
        atoms in the dictionary must have norm equal to one.

        Parameters
        ----------

        D: array_like or sparse matrix, shape= (n_atoms, n_features)
            The dictionary.

        penalty: a sub-class of Penalty class in penalty.py file, optional
            It is applied on the coefficients and it can be
            - L0Penalty
            - L1Penalty
            - L2Penalty
            - ElasticNetPenalty
            - GroupLassoPenalty
            - LInfPenalty

        non_negativity: bool, optional
            If true the coefficients are projected to the non-negative
            space.

        random_state: RandomState or int
            Seed to be use to initialise np.random.RandomState. If None each
            time RandomState is randomly initialised.

       """

    def __init__(self, D,  penalty=None, non_negativity=0, random_state=None):

        self.penalty = penalty
        self.non_negativity = non_negativity
        self.random_state = random_state
        self.X = None
        self.D = D
        self.C = None

    def fit(self, x, backtracking=0, n_iter=20000):

        """Function that fits the estimator on the matrices X and D.

        This function finds the coefficients with a proximal gradient
        algorithm. It computes the gradient on the reconstruction part of the
        functional and then applies the prox operator of the specified penalty.
        If the flat non-negative is set it project the rows of the matrices to
        the positive space.

        Parameters
        ----------

        x : array-like or sparse matrix shape =  (n_samples, n_features)
            The matrix to decompose.

        backtracking: bool, optional
            If True a procedure of backtracking is done on the step in order
            to avoid an increasing in the objective function.

        n_iter: int, optional
            Maximum number of iteration the minimization algorithm can
            perform.

        Returns
        -------
        self : object

        """

        # ______________ parameters control________________________________#
        x = check_array(x)
        self.D = check_array(self.D)
        self.X = x
        random_state = check_random_state(self.random_state)
        self.penalty = _check_penalty(self.penalty)

        logging.debug("Initializing coefficients..")

        # ________________optimization procedure____________________________#
        self.C = self._proximal_gradient_minimization(random_state,
                                                backtracking=backtracking,
                                                n_iter=n_iter)

        # __________________________final controls__________________________#
        logging.debug("Finished optimization")
        if self.non_negativity:
            assert (np.min(self.C) >= 0)

        return self

    def reconstruction_error(self):
        """
        Returns
        -------
        float:
            The reconstruction error for the current decomposition of the
            matrix. If no decomposition was run infinity is returned.
            A lower reconstruction error corresponds to a better approximation
            of the input data.
        """

        if self.X is None:
            return float("inf")
        return (np.linalg.norm(self.X - self.C.dot(self.D))/
                np.linalg.norm(self.X))

    def objective_function_value(self, x=None, d=None, c=None):
        """

        Parameters
        ----------
        x : array-like, shape=(n_samples, n_features)
            The matrix to be decomposed.

        d: array_like, shape=(n_atoms, n_features)
            The dictionary.

        c: array-like, shape=(n_samples, n_atoms)
            The matrix of coefficients

        If one of the three is None the internal decomposition is taken, if no
        decomposition is available NaN is returnes.

        Returns
        -------
        float:
            The value of the objective function.
        """
        if x is None:
            x = self.X
            if x is None:
                logging.info('You did not call the function fit. '
                             'Impossible to return an objective '
                             'function value')
                return float('NaN')

        if d is None:
            d = self.D
        if c is None:
            c = self.C

        reconstruction = np.linalg.norm(x - c.dot(d))**2
        penalty_coefficient = self.penalty.value(c)
        return reconstruction + penalty_coefficient

    def coefficients(self):
        """
        Returns
        -------

        C: array-like, shape = (n_samples, k)
            The learnt coefficients.

        """
        return self.C

    def score(self, *args):
        """
        Parameters
        ----------

        *args: optional
            Introduced for compatibility with sklearn GridSearchCV

        Returns
        -------
        float
            The score of the current decomposition. BIC value computed as
                 k(log(n_samples)-log(2\pi)) - 2*(||X - CD||)
            the highest is the score and better the decomposition is.
        """
        if self.X is None:
            return float("-inf")
        return - (np.log(self.X.shape[0])
                  + 2 * np.linalg.norm(self.X - self.C.dot(self.D)) ** 2)

    def _proximal_gradient_minimization(self, random_state, backtracking=0,
                                        n_iter=20000):
        x = self.X
        d = self.D
        n, p = x.shape
        k, _ = d.shape
        c = non_negative_projection(random_state.rand(n, k)*10-5,
                                    self.non_negativity)

        gamma = 1.1
        step = _step_lipschitz(d, gamma)
        iters = 20000
        epsilon = 1e-4
        objective = self.objective_function_value(d=d, c=c)

        c_old = c
        logging.debug("Starting optimization")
        for i in range(iters):
            gradient = (c.dot(d) - x).dot(d.T)

            if backtracking:
                c = self._update_with_backtracking(x, d, c, gradient, step)
            else:
                c = self._simple_update(c, gradient, step)

            new_objective = self.objective_function_value(d=d, c=c)
            difference_objective = new_objective - objective
            objective = new_objective
            difference_c = np.linalg.norm(c - c_old)
            c_old = c

            step = _step_lipschitz(d, gamma)

            logging.debug("Iteration: %i"%(i))
            logging.debug("Objective function: %f"%(objective))
            logging.debug("Difference between new objective and "
                          "previous one: %f"%(difference_objective))
            logging.debug("Difference between previous and new coefficients: "
                          "%f"%(difference_c))
            logging.debug("\n")

            assert ((not np.isnan(difference_objective)) and
                    (not np.isinf(difference_objective)) and
                    abs(difference_objective) < 1e+20)

            if (abs(difference_objective) <= epsilon and
                    difference_c <= epsilon):
                break

        return c

    def _simple_update(self, c, gradient, step):
        c = self.penalty. \
            apply_prox_operator(c - step * gradient, gamma=step)
        c = non_negative_projection(c, self.non_negativity)
        return c

    def _update_with_backtracking(self, x, d, c, gradient, step):
        c_0 = c
        error = 0.5 * np.linalg.norm(x - c_0.dot(d))**2
        sigma = 0.9
        for i in range(1, 1000):
            # compute new matrices
            c_1 = self.penalty. \
                apply_prox_operator(c_0 - step * gradient, gamma=step)
            c_1 = non_negative_projection(c_1, self.non_negativity)
            difference_c = np.linalg.norm(c_1 - c_0)

            first_part = self.objective_function_value(d=d, c=c_1)

            second_part = (error
                          + self.coeff_penalty.value(c_1)
                          + (difference_c*gradient).trace()
                          + step*sigma**i * difference_c)

            if first_part <= second_part:
                break

        return c_1


# ____________________________UTILITY FUNCTIONS_______________________________
def _step_lipschitz(d, gamma_c):
    step_c = max(0.0001, gamma_c * np.linalg.norm(d.T.dot(d)))
    return 1/step_c








from __future__ import print_function, division

import matplotlib.pyplot as plt
import numpy as np
from scipy import signal
from sklearn.utils import check_random_state


def group_lasso_dataset_generator(n_samples=100, n_features=100,
                                  gaussian_noise=0.5, random_state=None):
    """
        Generates synthetic data for group lasso tests.

        This function generates a matrix generated from 7 basic atoms, grouped
        as [0, 1, 3], [2, 4, 5], linearly combined with random weights.
        A certain level of gaussian noise is added to the signal.

        Parameters
        ----------
         n_samples: int, optional
        Number of samples for the output matrix.

        n_features: int, optional
            Number of features the output matrix must have.

        gaussian_noise: float, optional
            The level of noise to add to the synthetic data.

        random_state: RandomState or int, optional
            RandomState or seed used to generate RandomState for the
            reproducibility of data. If None each time RandomState is randomly
            initialised.

        Returns
        -------
        array_like, shape=(n_samples, n_features)
            Generated matrix of data

        array_like, shape=(n_samples, 7)
            Coefficients

        array_like, shape=(7, n_features)
            Dictionary
        """
    rnd = check_random_state(random_state)
    number_of_atoms = 6

    atoms = np.empty([n_features, number_of_atoms])
    t = np.linspace(0, 1, n_features)
    atoms[:, 0] = signal.sawtooth(2 * np.pi * 5 * t)
    atoms[:, 1] = np.sin(2 * np.pi * t)
    atoms[:, 2] = np.sin(2 * np.pi * t - 15)
    atoms[:, 3] = signal.gaussian(n_features, 5)
    atoms[:, 4] = signal.square(2 * np.pi * 5 * t)
    atoms[:, 5] = np.abs(np.sin(2 * np.pi * t))

    groups = [[0, 1, 3], [2, 4, 5]]

    signals = np.empty((n_samples, n_features))
    coefficients = np.zeros((n_samples, number_of_atoms))
    for i in range(n_samples // 2):
        coeffs = rnd.random_sample(len(groups[0])) * 10
        coefficients[i, groups[0]] = coeffs

    for i in range(n_samples // 2, n_samples):
        coeffs = rnd.random_sample(len(groups[1])) * 10
        coefficients[i, groups[1]] = coeffs

    signals = coefficients.dot(atoms.T)

    return signals, coefficients, atoms.T


def sparse_signal_generator(n_samples, n_features, frequencies,
                            support_atoms, shift=True):
    # TODO: sistemare questa documentazione
    """ The following function generates signals using sawtooth and sin

    Parameters
    -------------------
    n_samples : int
        number of signals to be generated

    n_features : int
        length of the time series (number of points)

    frequencies :
        number of frequencies (to be used for the def of atoms)

    support_atoms:
        qualcosa

    shift :
        if true shifted atoms, else fixed

    Returns
    -------------------
    multichannel_matrix : np.array(n_features, n_samples)
         matrix of signals
    atoms_matrix :  np.array(n_features, number_of_atoms)
         matrix of signals
    """

    f_array = np.linspace(4. / n_features, 40. / n_features, frequencies)
    atom_shape = 2

    if shift:
        n_shifts = n_features - support_atoms
    else:
        n_shifts = 1

    n_atoms = frequencies * atom_shape * n_shifts
    _low = int(0.4 * n_atoms)
    _high = int(0.7 * n_atoms)
    selected_atoms = np.random.randint(low=_low, high=_high, size=(10,))
    atoms = np.zeros((n_features, n_atoms))
    time_vector = np.arange(support_atoms)
    diff_supp = n_features - support_atoms

    for i in range(frequencies):
        temp1 = np.sin(f_array[i] * time_vector)
        temp2 = signal.sawtooth(f_array[i] * time_vector)
        norm1 = np.linalg.norm(np.pad(temp1, (0, diff_supp), mode='constant'))
        norm2 = np.linalg.norm(np.pad(temp2, (0, diff_supp), mode='constant'))
        for j in range(n_shifts):
            atoms[:, i * n_shifts + j] = np.pad(temp1, (j, diff_supp - j),
                                                mode='constant') / norm1
            atoms[:, i * n_shifts + j + frequencies * n_shifts] = \
                np.pad(temp2, (j, diff_supp - j), mode='constant') / norm2

    multichannel_signal = np.zeros((n_features, n_samples))
    for i in range(n_samples):
        random_atoms = np.random.choice(selected_atoms, size=5)
        weight = 10 * np.random.randn(5, )
        multichannel_signal[:, i] = np.dot(atoms[:, random_atoms], weight)

    np.save('signal_gen', multichannel_signal)
    np.save('atom_gen', atoms)

    return multichannel_signal, atoms


def synthetic_data_non_negative(gaussian_noise=1, random_state=None):
    """
    Generates synthetic non-negative data for dictionary learning tests.

    This function generates a matrix generated from 7 basic atoms linearly
    combined with random weights sparse over the atoms. A certain level of
    gaussian noise is added to the signal.

    Parameters
    ----------
    gaussian_noise: float, optional
        The level of noise to add to the synthetic data.

    random_state: RandomState or int, optional
        RandomState or seed used to generate RandomState for the
        reproducibility of data. If None each time RandomState is randomly
        initialised.

    Returns
    -------
    array_like, shape=(80, 96)
        Generated matrix of data

    array_like, shape=(80, 7)
        Coefficients

    array_like, shape=(7, 96)
        Dictionary
    """

    number_of_features = 96
    number_of_samples = 80
    number_of_atoms = 7
    rnd = check_random_state(random_state)

    atoms = np.empty([number_of_features, number_of_atoms])
    atoms[:, 0] = np.transpose(
        np.concatenate((np.ones([30, 1]), np.zeros([66, 1]))))
    atoms[:, 1] = np.transpose(
        np.concatenate((np.zeros([60, 1]), np.ones([36, 1]))))
    atoms[:, 2] = np.transpose(np.concatenate(
        (np.zeros([24, 1]), np.ones([30, 1]), np.zeros([42, 1]))))
    atoms[:, 3] = signal.gaussian(96, 5)
    atoms[:, 4] = np.transpose(np.concatenate((np.zeros([17, 1]),
                                               np.ones([15, 1]),
                                               np.zeros([30, 1]),
                                               np.ones([24, 1]),
                                               np.zeros([10, 1]))))
    atoms[:, 5] = np.roll(signal.gaussian(96, 5), 30)
    atoms[:, 6] = signal.gaussian(96, 8)
    atoms[0:50, 6] = 0

    sums = np.sum(atoms, axis=0)
    atoms = atoms / sums

    # create sparse coefficients
    coefficients = np.zeros([number_of_atoms, number_of_samples])
    for i in range(0, number_of_samples):
        number_of_nonzero_elements = rnd.randint(2, 4)
        indices = rnd.choice(range(0, 7), number_of_nonzero_elements,
                             replace=False)
        coeffs = rnd.random_sample(number_of_nonzero_elements) * 100
        coefficients[indices, i] = coeffs

    # create matrix
    v = np.dot(atoms, coefficients)

    # add noise
    v_tilde = v + np.random.normal(0, gaussian_noise,
                                   (number_of_features, number_of_samples))
    v_tilde[np.where(v_tilde < 0)] = 0

    return v_tilde.T, coefficients.T, atoms.T


def synthetic_data_negative(n_samples=100, n_features=60,
                            gaussian_noise=1, random_state=None):
    """
    Generates synthetic data for dictionary learning tests.

    This function generates a matrix generated from 10 basic atoms linearly
    combined with random weights sparse over the atoms. A certain level of
    gaussian noise is added to the signal.

    Parameters
    ----------
    n_samples: int, optional
        Number of samples for the output matrix.

    n_features: int, optional
        Number of features the output matrix must have.

    gaussian_noise: float, optional
        The level of noise to add to the synthetic data.

    random_state: RandomState or int, optional
        RandomState or seed used to generate RandomState for the
        reproducibility of data. If None each time RandomState is randomly
        initialised.

    Returns
    -------
    array_like, shape=(n_samples, number_of_features)
        Generated matrix of data

    array_like, shape=(n_samples, 5)
        Coefficients

    array_like, shape=(5, n_features)
        Dictionary
    """

    plt.close("all")
    n_atoms = 5
    rnd = check_random_state(random_state)

    atoms = np.empty([n_features, n_atoms])

    t = np.linspace(0, 1, n_features)
    atoms[:, 0] = np.sin(2 * np.pi * t)
    atoms[:, 1] = np.sin(2 * np.pi * t - 15)
    atoms[:, 2] = np.abs(np.sin(2 * np.pi * t))
    z = signal.gausspulse(t - 0.5, fc=5, retquad=True, retenv=True)
    atoms[:, 3] = z[2]
    atoms[:, 4] = np.roll(np.sign(z[2] - 0.5), 10)

    # create sparse coefficients
    coefficients = np.zeros([n_atoms, n_samples])
    for i in range(0, n_samples):
        number_of_nonzero_elements = rnd.randint(2, 4)
        indices = rnd.choice(range(0, n_atoms),
                             number_of_nonzero_elements,
                             replace=False)
        coeffs = rnd.random_sample(number_of_nonzero_elements) * 10
        coefficients[indices, i] = coeffs

    # create matrix
    v = np.dot(atoms, coefficients)

    # add noise
    v_tilde = v + np.random.normal(0, gaussian_noise,
                                   (n_features, n_samples))
    return v_tilde.T, coefficients.T, atoms.T



from __future__ import print_function, division

import logging

from sklearn.model_selection._split import BaseShuffleSplit
from sklearn.model_selection._split import _validate_shuffle_split
from sklearn.utils import check_random_state
import numpy as np


def non_negative_projection(x, nn, matrix_type=None):
    if matrix_type is None and nn:
        np.maximum(x, 0, out=x)
    if nn == 'both' or nn == matrix_type:
        np.maximum(x, 0, out=x)
    return x


def _check_non_negativity(nn, x):
    if nn == 'both':
        if np.min(x) < 0:
            logging.error('The matrix of signals to decompose has '
                          'negative numbers, impossible to use non-negative'
                          ' matrix factorization.')
            raise ValueError('The matrix of signals to decompose has '
                          'negative numbers, impossible to use non-negative'
                          ' matrix factorization.')
    if not (nn == 'both' or nn == 'none' or nn == 'coeff'):
        logging.error("Unknown option for non_negativy, please use"
                      "one of {'none', 'both', 'coeff'}.")
        raise ValueError("Unknown option for non_negativy, please use"
                      "one of {'none', 'both', 'coeff'}.")


def _compute_clusters_and_silhouettes(Ds, Cs):
    all_dictionaries, all_coefficients, centroids, clusters = \
        _clustering(Ds, Cs)
    silhouettes = _silhouette(clusters)

    _sums = np.zeros_like(all_coefficients[0])

    for i in range(0, len(all_coefficients)):
        _sums = _sums + all_coefficients[i]
    mean_coeffs = _sums / len(all_coefficients)

    return all_dictionaries, all_coefficients, centroids, \
           mean_coeffs, np.mean(silhouettes)


def _clustering(Ds, Cs):
    # take the last element (not clustered yet)
    iterations = len(Ds)
    index_last = iterations - 1
    D = Ds[index_last]
    C = Cs[index_last]

    # if first iteration the dictionary is already the cluster
    if iterations == 1:
        return Ds, Cs, D, D

    # take dimensions
    number_of_atoms, number_of_features = D.shape

    # recompute centroids and clusters
    clusters = []
    sums = np.zeros((number_of_atoms, number_of_features))
    for k in range(number_of_atoms):
        cluster = []
        for d in range(iterations - 1):
            sums[k, :] = sums[k, :] + Ds[d][k, :]
            cluster.append(Ds[d][k, :])
        clusters.append(cluster)

    centroids = sums / (iterations - 1)

    # find the nearest element in D to each centroids
    new_order_for_atoms = np.zeros(number_of_atoms)
    used_atoms = []
    for c in range(number_of_atoms):
        distances = compute_correlation(centroids[c, :], D, used_atoms)
        index = np.argmax(distances)
        new_order_for_atoms[c] = index
        centroids[c, :] = (sums[c, :] + D[index, :]) / iterations
        clusters[c].append(D[index,])
        used_atoms.append(index)

    # order D_last and H_last w.r.t. ordered_D
    new_D = np.zeros_like(D)
    new_C = np.zeros_like(C)
    for i in range(number_of_atoms):
        new_D[i, :] = D[int(new_order_for_atoms[i]), :]
        new_C[:, i] = C[:, int(new_order_for_atoms[i])]

    Ds[index_last] = new_D
    Cs[index_last] = new_C

    return Ds, Cs, centroids, clusters


def _silhouette(clusters):
    silhouettes = np.zeros(len(clusters))
    for c in range(len(clusters)):
        C = clusters[c]
        others = [x for i, x in enumerate(clusters) if i != c]
        silhouettes[c] = _single_cluster_silhouette(C, others)
    return silhouettes


def _single_cluster_silhouette(A, others):
    single_point_silhouette = np.zeros(len(A))
    for i in range(len(A)):
        a_i = _average_distance(A[i], A)

        d_i = np.zeros(len(others))
        for c in range(len(others)):
            d_i[c] = _average_distance(A[i], others[c])
        b_i = np.max(d_i)

        if a_i > b_i:
            single_point_silhouette[i] = 1 - b_i / a_i
        elif a_i == b_i:
            single_point_silhouette[i] = 0
        else:
            single_point_silhouette[i] = a_i / b_i - 1
    return np.mean(single_point_silhouette)


def _average_distance(point, cluster):
    sum_of_distances = 0
    for new_point in cluster:
        sum_of_distances = sum_of_distances + np.dot(point, new_point)
    return sum_of_distances / len(cluster)


def compute_correlation(x, set_to_compare, not_to_consider):
    """
    Computes correlations between x and the set of other signals.

    Given a signal x it computes the correlations with all the other signals
    in set_to_compare. The correlation of the signals whose indices are in
    not_to_consider will be -inf.

    Parameters
    ----------
    x: array-like, shape=(1, n_features)
        Signal to compare.

    set_to_compare: array-like or list, shape=(n_samples, n_features)
        Matrix of signals on which compute correlation.

    not_to_consider: list of int
        Indices of the matrix set_to_compare that do not have to be
        considered. The correlation for this signals will be set to -inf.

    Returns
    -------
    array-like, shape=(n_samples)
        The correlations of all the signals in set_to_compare with x.
    """

    set_to_compare = np.array(set_to_compare)
    correlations = np.zeros(set_to_compare.shape[0])
    for a in range(set_to_compare.shape[0]):
        if a in not_to_consider:
            correlations[a] = -float("inf")
        else:
            vector = set_to_compare[a, :]
            correlations[a] = (np.dot(x, vector.T))
    return correlations


class MonteCarloBootstrap(BaseShuffleSplit):
    """
    Montecarlo sampling

    It does random sampling with replacement where if the dataset has
    dimensions (n, p) it samples n random elements from the dataset where some
    of them can be repeated multiple times.

    Parameters
    ----------
    n_splits : int (default 10)
        Number of re-shuffling & splitting iterations.

    test_size : float, int, or None, default 0.1
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the test split. If
        int, represents the absolute number of test samples. If None,
        the value is automatically set to the complement of the train size.

    train_size : float, int, or None (default is None)
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the train split. If
        int, represents the absolute number of train samples. If None,
        the value is automatically set to the complement of the test size.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Examples
    --------
    from dalila.utils import MonteCarloBootstrap
    X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
    rs = MonteCarloBoostrap(n_splits=3, test_size=.25, random_state=0)
    rs.get_n_splits(X)

    for train_index, test_index in rs.split(X):
        print("TRAIN:", train_index, "TEST:", test_index)
    """

    def _iter_indices(self, X, y=None, groups=None):
        n_samples = X.shape[0]
        n_train, n_test = _validate_shuffle_split(n_samples, self.test_size,
                                                  self.train_size)
        rng = check_random_state(self.random_state)
        for i in range(self.n_splits):
            # random partition
            ind_train = rng.randint(0, high=X.shape[0],
                                    size=n_train)
            ind_test = list(set(np.arange(0, X.shape[0])) -
                            set(np.unique(ind_train)))
            yield ind_train, ind_test


from __future__ import print_function
import sys
from setuptools import setup, find_packages

with open('requirements.txt') as f:
    INSTALL_REQUIRES = [l.strip() for l in f.readlines() if l]

setup(name='dalila',
      version='1.0.0',
      description='A library for dictionary learning decomposition',
      long_description='A library for the decomposition of a matrix of signals.'
                       'It includes sparse coding algorithm and parameter '
                       'research procedures.',
      url='https://github.com/slipguru/dalila',
      author='Veronica Tozzo',
      author_email='veronica.tozzo@dibris.unige.it',
      license='FreeBSD',
      classifiers={
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Developers',
          'Programming Language :: Python',
          'License :: OSI Approved :: BSD License',
          'Topic :: Software Development',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Operating System :: POSIX',
          'Operating System :: Unix'},
      packages=find_packages(),
      install_requires=INSTALL_REQUIRES,
      )




import matplotlib.pyplot as plt
import numpy as np


def plot_dictionary_atoms(dictionary):
    """
    It plots the atoms composing the dictionary.

    Parameters
    ----------
    dictionary: array-like, shape=(n_atoms, n_features)

    """

    for r in range(0, dictionary.shape[0]):
        plt.figure()
        plt.plot(dictionary[r, :])
    plt.show()


def plot_atoms_as_histograms(dictionary):
    """
    It plots the atoms composing the dictionary as histograms.

    Parameters
    ----------
    dictionary: array_like, shape=(n_atoms, n_features)
    """
    for i in range(0, dictionary.shape[0]):
        fig = plt.figure()
        fig.canvas.set_window_title(str(i+1) + " atom")
        length = len(dictionary[i, :])
        x = np.asarray(range(0, length))
        w = dictionary[i, :]
        plt.hist(x, bins=length, weights=w)
        plt.xlim((0, dictionary.shape[1]))
    plt.show()

